#!/usr/bin/env python3

from sys import stderr, path
from os import getenv
from os.path import dirname, abspath, join
from dotenv import load_dotenv
import getpass
from types import GeneratorType

import dill

from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor
from twisted.internet.endpoints import TCP4ServerEndpoint


class AchillesServer(LineReceiver):
    MAX_LENGTH = 999999999999999999999999999999999

    def __init__(self, factory):
        self.factory = factory
        print("Starting factory:", self.factory)
        self.HOST = (
            self.factory.HOST
        )  # The server's hostname or IP address - standard loopback interface address (localhost)
        self.PORT = (
            self.factory.PORT
        )  # Port to listen on (non-privileged ports are > 1023)
        self.CPU_COUNT = 0
        self.IP = ""
        self.DATETIME_CONNECTED = ""
        self.AUTHENTICATED = False
        self.CLIENT_ID = self.factory.totalProtocols

    def connectionMade(self):
        # print("Starting connection #:", self.factory.numProtocols)
        self.factory.numProtocols = self.factory.numProtocols + 1
        self.factory.totalProtocols = self.factory.totalProtocols + 1
        # print("Number of connections:", self.factory.numProtocols)
        self.factory.clients.append(self)
        packet = dill.dumps(
            {
                "GREETING": f"Welcome! There are currently {self.factory.numProtocols} open connections.\n",
                "CLIENT_ID": self.CLIENT_ID,
            }
        )
        self.sendLine(packet)

    def connectionLost(self, reason):
        self.factory.numProtocols = self.factory.numProtocols - 1
        print(f"Connection lost: {reason}")
        print(f"{self.factory.numProtocols} clients connected.")
        self.factory.clients.remove(self)

    def lineReceived(self, data):
        data = dill.loads(data)
        # print("RECEIVED:", data)
        if "USERNAME" in data and "SECRET_KEY" in data:
            if (
                data["USERNAME"] == self.factory.USERNAME
                and data["SECRET_KEY"] == self.factory.SECRET_KEY
            ):
                # The user is authenticated to distribute commands.
                self.AUTHENTICATED = True
                self.IP = data["IP"]
                self.CPU_COUNT = data["CPU_COUNT"]
                self.DATETIME_CONNECTED = data["DATETIME_CONNECTED"]
                self.sendLine(dill.dumps({"AUTHENTICATED": self.AUTHENTICATED}))
                print(f"User {data['USERNAME']} is authenticated.")
                self.factory.achilles_controller = self

                # for client in self.factory.clients:
                # print(client.__dict__)
            else:
                self.sendLine(dill.dumps({"AUTHENTICATED": self.AUTHENTICATED}))
                stderr.write(
                    "This USERNAME and SECRET_KEY cannot be authenticated. Closing connection."
                )
                self.transport.loseConnection()
                # for client in self.factory.clients:
                # print(client.__dict__)
        elif "IP" in data and "CPU_COUNT" in data and self.AUTHENTICATED is False:
            # print("Clients:", self.factory.clients)
            self.IP = data["IP"]
            self.CPU_COUNT = data["CPU_COUNT"]
            self.DATETIME_CONNECTED = data["DATETIME_CONNECTED"]
            self.CLIENT_ID = data["CLIENT_ID"]
            # for client in self.factory.clients:
            # print(client.__dict__)
        elif self.AUTHENTICATED is True and "FUNC" in data:
            func = data["FUNC"]
            args = data["ARGS"]
            args_path = data["ARGS_PATH"]
            modules = data["MODULES"]
            callback = data["CALLBACK"]
            reducer = data["REDUCER"]
            group = data["GROUP"]
            response_mode = data["RESPONSE_MODE"]
            chunksize = data["CHUNKSIZE"]
            self.factory.response_mode = response_mode
            self.factory.chunksize = chunksize

            self.startJob(func, args, args_path, modules, callback, reducer, group)
        elif self.AUTHENTICATED is False and "READY" in data:
            # print(f"CLIENT STATUS: {data}")
            pass

        elif self.AUTHENTICATED is True and "VERIFY" in data:
            self.proceedWithJob()

        elif "RESULT" in data:
            self.argsBufferDel(data, self.factory.args_buffer)
            # print("RESULTS PACKET:", data)
            if self.factory.response_mode == "OBJECT":
                self.factory.results.append(data)
                try:
                    loadBalance(self.factory.args, self)
                except StopIteration:
                    self.handleStopIteration(response_mode="OBJECT")

            elif (
                self.factory.response_mode == "STREAM"
                or self.factory.response_mode == "SQLITE"
            ):
                self.factory.achilles_controller.sendLine(dill.dumps(data))
                try:
                    loadBalance(self.factory.args, self)
                except StopIteration:
                    self.handleStopIteration(response_mode="STREAM")

        elif "GET_CLUSTER_STATUS" in data and self.AUTHENTICATED is True:
            packet = {"CLUSTER_STATUS": True}
            for client in self.factory.clients:
                packet[str(client.CLIENT_ID)] = {}
                packet[str(client.CLIENT_ID)]["CLIENT_ID"] = str(client.CLIENT_ID)
                packet[str(client.CLIENT_ID)]["HOST"] = str(client.HOST)
                packet[str(client.CLIENT_ID)]["PORT"] = str(client.PORT)
                packet[str(client.CLIENT_ID)]["CPU_COUNT"] = str(client.CPU_COUNT)
                packet[str(client.CLIENT_ID)]["IP"] = str(client.IP)
                packet[str(client.CLIENT_ID)]["DATETIME_CONNECTED"] = str(
                    client.DATETIME_CONNECTED
                )
                packet[str(client.CLIENT_ID)]["AUTHENTICATED"] = str(
                    client.AUTHENTICATED
                )

            self.factory.achilles_controller.sendLine(dill.dumps(packet))

        elif "KILL_CLUSTER" in data and self.AUTHENTICATED is True:
            for client in self.factory.clients:
                client.sendLine(dill.dumps({"KILL_NODE": "KILL_NODE"}))

            for client in self.factory.clients:
                client.transport.loseConnection()

        elif "KILLED_CLUSTER" in data:
            reactor.stop()

        else:
            print("UNEXPECTED", data)

    def startJob(
        self,
        func,
        args=(),
        args_path="",
        modules=None,
        callback=None,
        reducer=None,
        group="",
    ):
        # Here is where the magic happens. Hungry consumers - feed them once and they keep
        # asking for more until the args are exhausted.

        # Flush settings in case another job has already been completed in this lifecycle.
        self.factory.args_path = ""
        self.factory.results = []
        self.factory.args_buffer = {}
        self.factory.args_counter = 0
        self.factory.callback = callback
        self.factory.reducer = reducer

        # print("MAP FUNC:", func)
        # print("MAP ARGS:", args)
        ip_list = []
        ip_map = []
        workers_list = []
        cpu_total = 0
        try:
            self.factory.args = iter(args(args_path))
        except TypeError:
            try:
                self.factory.args = iter(args)
            except TypeError:
                self.factory.args = iter(args())
        for client in self.factory.clients:
            if client.IP not in ip_list:
                ip_list.append(client.IP)
                ip_map.append((client.IP, client.CPU_COUNT))
                # Int may be unnecessary - check data type.
                cpu_total += int(client.CPU_COUNT)
            if client.AUTHENTICATED is False:
                workers_list.append(client)
        print("IP LIST:", ip_list)
        print("CPU TOTAL:", cpu_total)
        print("CLIENTS CONNECTED:", self.factory.numProtocols)
        self.factory.ipMap = ip_map
        self.factory.workers = workers_list
        self.factory.lastCounter = len(self.factory.workers) - 1
        for client in self.factory.workers:
            client.sendLine(
                dill.dumps(
                    {
                        "START_JOB": True,
                        "FUNC": func,
                        "CALLBACK": callback,
                        "REDUCER": reducer,
                    }
                )
            )
        self.factory.achilles_controller.sendLine(dill.dumps({"PROCEED": True}))

    def proceedWithJob(self):
        for worker in self.factory.workers:
            try:
                loadBalance(self.factory.args, worker)
            except StopIteration:
                self.handleStopIteration(self.factory.response_mode)

    def argsBufferDel(self, result_packet, args_buffer):
        del args_buffer[result_packet["ARGS_COUNTER"]]

    def reassignArgFromBuffer(self, args_buffer):
        args_counter, args_packet = next(iter(args_buffer.items()))
        packet = dill.dumps(args_packet)
        self.sendLine(packet)

    def handleStopIteration(self, response_mode):
        if response_mode == "STREAM":
            if len(self.factory.workers) > 1:
                if self.factory.lastCounter == 0:
                    try:
                        self.reassignArgFromBuffer(self.factory.args_buffer)
                    except StopIteration:
                        print(
                            "Final results packet has been transmitted to the achilles_controller."
                        )
                        self.factory.achilles_controller.sendLine(
                            dill.dumps({"JOB_FINISHED": True})
                        )
                else:
                    self.factory.lastCounter = self.factory.lastCounter - 1
            elif len(self.factory.workers) == 1:
                try:
                    self.reassignArgFromBuffer(self.factory.args_buffer)
                except StopIteration:
                    print(
                        "Final results packet has been transmitted to the achilles_controller."
                    )
                    self.factory.achilles_controller.sendLine(
                        dill.dumps({"JOB_FINISHED": True})
                    )

        elif response_mode == "OBJECT":
            if len(self.factory.workers) > 1:
                if self.factory.lastCounter == 0:
                    try:
                        self.reassignArgFromBuffer(self.factory.args_buffer)
                    except StopIteration:
                        final_result = self.factory.gatherResults(self.factory)
                        self.factory.achilles_controller.sendLine(
                            dill.dumps({"FINAL_RESULT": final_result})
                        )
                        print(
                            "Final results packet has been transmitted to the achilles_controller."
                        )

                else:
                    self.factory.lastCounter = self.factory.lastCounter - 1
            elif len(self.factory.workers) == 1:
                try:
                    self.reassignArgFromBuffer(self.factory.args_buffer)
                except StopIteration:
                    final_result = self.factory.gatherResults(self.factory)
                    self.factory.achilles_controller.sendLine(
                        dill.dumps({"FINAL_RESULT": final_result})
                    )
                    print(
                        "Final results packet has been transmitted to the achilles_controller."
                    )


def loadBalance(args_iterable, worker):
    test_arg = next(args_iterable)
    if type(test_arg) is list:
        packet = {"ARG": test_arg, "ARGS_COUNTER": int(worker.factory.args_counter)}
        argsBufferAdd(packet, worker.factory.args_buffer)

        worker.sendLine(dill.dumps(packet))
        """print(
            f"Packet with arg {worker.factory.args_counter} sent to {worker.CLIENT_ID}"
        )"""
        worker.factory.args_counter = worker.factory.args_counter + 1
    else:
        cpu_count = int(worker.CPU_COUNT)
        args_counter = worker.factory.args_counter
        packet = {"ARG": [test_arg], "ARGS_COUNTER": args_counter}

        worker.factory.args_counter = worker.factory.args_counter + 1

        balancer = worker.factory.chunksize * cpu_count - 1
        for i in range(balancer):
            try:
                packet["ARG"].append(next(worker.factory.args))
                worker.factory.args_counter = worker.factory.args_counter + 1
            except StopIteration:
                break

        argsBufferAdd(packet, worker.factory.args_buffer)
        worker.sendLine(dill.dumps(packet))
        # print(f"Packet with arg {args_counter} sent to {worker.CLIENT_ID}")


def argsBufferAdd(arg_packet, args_buffer):
    args_buffer = args_buffer
    args_buffer[arg_packet["ARGS_COUNTER"]] = arg_packet
    return args_buffer


class AchillesServerFactory(Factory):

    protocol = AchillesServer
    numProtocols = 0
    totalProtocols = 0
    clients = []
    workers = []
    achilles_controller = None
    args = None
    args_buffer = {}
    args_path = ""
    args_counter = 0
    results = []
    lastCounter = 0
    ipMap = []
    response_mode = None
    HOST = ""
    PORT = 0
    USERNAME = ""
    SECRET_KEY = ""

    def __init__(self, host, port, username, secret_key):
        AchillesServerFactory.HOST = host
        AchillesServerFactory.PORT = port
        AchillesServerFactory.USERNAME = username
        AchillesServerFactory.SECRET_KEY = secret_key

    def buildProtocol(self, addr):
        return AchillesServer(factory=AchillesServerFactory)

    def gatherResults(self):
        final_results = []
        self.results = sorted(self.results, key=lambda k: k["ARGS_COUNTER"])
        for result in self.results:
            final_results.append(result["RESULT"])
        return final_results


def runAchillesServer(host=None, port=None, username=None, secret_key=None):
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass
    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )

                achilles_function_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\"
                )
                path.append(achilles_function_path)

            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")

                achilles_function_path = abspath(dirname(__file__))
                path.append(achilles_function_path)
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")
            username = getenv("USERNAME")
            secret_key = getenv("SECRET_KEY")

        except BaseException as e:
            print(f"No .env configuration file found ({e})...")
            host, port, username, secret_key = genConfig()

    endpoint = TCP4ServerEndpoint(reactor, port)
    endpoint.listen(AchillesServerFactory(host, port, username, secret_key))
    print(f"ALERT: achilles_server initiated at {host}:{port}\n")
    print("Listening for connections...")
    reactor.run()


def genConfig(host=None, port=None, username=None, secret_key=None):
    if __name__ != "__main__":
        import achilles

        dotenv_path = abspath(dirname(achilles.__file__)) + "\\lineReceiver\\"
    else:
        basedir = abspath(dirname(__file__))
        dotenv_path = join(basedir, ".env")

    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass
    else:
        host = input("Enter HOST IP address:\t")
        port = int(input("Enter host PORT to listen on:\t"))
        username = input("Enter USERNAME to require for authentication:\t")
        secret_key = getpass.getpass(
            "Enter SECRET_KEY to require for authentication:\t"
        )
    with open(dotenv_path + ".env", "w") as config_file:
        config_file.writelines(f"HOST={host}\n")
        config_file.writelines(f"PORT={port}\n")
        config_file.writelines(f"USERNAME='{username}'\n")
        config_file.writelines(f"SECRET_KEY='{secret_key}'\n")
        config_file.close()
        print(
            f"Successfully generated .env configuration file at {dotenv_path}. Use achilles_server.genConfig() to overwrite."
        )
    return host, port, username, secret_key


if __name__ == "__main__":
    runAchillesServer()



#!/usr/bin/env python3

from achilles.lineReceiver.achilles_controller import AchillesController

from dotenv import load_dotenv
from os import getenv
from os.path import abspath, dirname, join
import getpass
import time

from twisted.internet.endpoints import TCP4ClientEndpoint, connectProtocol
from twisted.internet import reactor
import multiprocess


def runAchillesController(
    achilles_function=None,
    achilles_args=None,
    achilles_callback=None,
    achilles_reducer=None,
    response_mode="OBJECT",
    host=None,
    port=None,
    username=None,
    secret_key=None,
    globals_dict=None,
    chunksize=1,
    command=None,
    command_verified=False,
    TCP4ClientEndpoint=TCP4ClientEndpoint,
    reactor=reactor,
    connectProtocol=connectProtocol,
    AchillesController=AchillesController,
    abspath=abspath,
    dirname=dirname,
):
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass
    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )
            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")
            username = getenv("USERNAME")
            secret_key = getenv("SECRET_KEY")

        except BaseException as e:
            try:
                host, port, username, secret_key = genConfig()
            except NameError:
                host = host
                port = port
                username = username
                secret_key = secret_key
                # print(f'Passed host {host}, port {port}, username {username}, secret_key {secret_key} assignment.')

    endpoint = TCP4ClientEndpoint(reactor, host, port)
    d = connectProtocol(
        endpoint,
        AchillesController(
            host,
            port,
            username,
            secret_key,
            achilles_function,
            achilles_args,
            achilles_callback,
            achilles_reducer,
            response_mode,
            globals_dict,
            chunksize,
            command,
            command_verified,
        ),
    )

    reactor.run()


def genConfig(host=None, port=None, username=None, secret_key=None):
    if __name__ != "__main__":
        import achilles

        dotenv_path = abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
    else:
        basedir = abspath(dirname(__file__))
        dotenv_path = join(basedir, ".env")
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass
    else:
        host = input("Enter HOST IP address:\t")
        port = int(input("Enter HOST port to listen on:\t"))
        username = input("Enter USERNAME to require for authentication:\t")
        secret_key = getpass.getpass(
            "Enter SECRET_KEY to require for authentication:\t"
        )
    with open(dotenv_path, "w") as config_file:
        config_file.writelines(f"HOST={host}\n")
        config_file.writelines(f"PORT={port}\n")
        config_file.writelines(f"USERNAME='{username}'\n")
        config_file.writelines(f"SECRET_KEY='{secret_key}'\n")
        config_file.close()
        print(
            f"Successfully generated .env configuration file at {dotenv_path}. Use achilles_controller.genConfig() to overwrite."
        )
    return host, port, username, secret_key


def imap_unordered(
    func,
    args,
    callback=None,
    reducer=None,
    response_mode="STREAM",
    host=None,
    port=None,
    username=None,
    secret_key=None,
    chunksize=1,
    globals_dict=None,
    runAchillesController=runAchillesController,
    abspath=abspath,
    dirname=dirname,
    multiprocess=multiprocess,
):

    if globals_dict is None:
        manager = multiprocess.Manager()

        globals_dict = {"OUTPUT_QUEUE": manager.Queue()}
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass

    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )
            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")
            username = getenv("USERNAME")
            secret_key = getenv("SECRET_KEY")
        except (KeyError, NameError):
            host = host
            port = port
            username = username
            secret_key = secret_key

    a = multiprocess.Process(
        target=runAchillesController,
        args=(
            func,
            args,
            callback,
            reducer,
            response_mode,
            host,
            port,
            username,
            secret_key,
            globals_dict,
            chunksize,
        ),
    )

    a.start()
    while True:
        result = globals_dict["OUTPUT_QUEUE"].get()
        if result != "JOB_FINISHED":
            yield result
        else:
            break

    a.terminate()


def imap(
    func,
    args,
    callback=None,
    response_mode="STREAM",
    host=None,
    port=None,
    username=None,
    secret_key=None,
    chunksize=1,
    globals_dict=None,
    runAchillesController=runAchillesController,
    abspath=abspath,
    dirname=dirname,
    multiprocess=multiprocess,
):

    if globals_dict is None:
        manager = multiprocess.Manager()

        globals_dict = {"OUTPUT_QUEUE": manager.Queue()}
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass
    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )
            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")
            username = getenv("USERNAME")
            secret_key = getenv("SECRET_KEY")
        except (KeyError, NameError):
            host = host
            port = port
            username = username
            secret_key = secret_key

    reducer = None

    a = multiprocess.Process(
        target=runAchillesController,
        args=(
            func,
            args,
            callback,
            reducer,
            response_mode,
            host,
            port,
            username,
            secret_key,
            globals_dict,
            chunksize,
        ),
    )

    a.start()
    sent_first_arg = False
    expected_args_counter = 0
    result_buffer = []
    try:
        test_arg = next(args())
        test_arg_type = type(test_arg)
    except TypeError:
        test_arg = next(iter(args))
        test_arg_type = type(test_arg)
    while True:
        result = globals_dict["OUTPUT_QUEUE"].get()
        if result != "JOB_FINISHED":
            if test_arg_type is not list:
                if sent_first_arg is False:
                    if result["ARGS_COUNTER"] == 0:
                        sent_first_arg = True
                        expected_args_counter = expected_args_counter + len(
                            result["RESULT"]
                        )
                        yield result
                    else:
                        # print("NOT YET:", result)
                        result_buffer.append(result)
                        continue
                else:
                    if result["ARGS_COUNTER"] == expected_args_counter:
                        expected_args_counter = expected_args_counter + len(
                            result["RESULT"]
                        )
                        yield result
                    else:
                        returned = getResult(result_buffer, expected_args_counter)
                        result_buffer.append(result)
                        if (
                            returned is not None
                            and returned["ARGS_COUNTER"] == expected_args_counter
                        ):
                            expected_args_counter = expected_args_counter + len(
                                result["RESULT"]
                            )
                            yield returned

                        else:
                            continue

            else:
                if sent_first_arg is False:
                    if result["ARGS_COUNTER"] == 0:
                        sent_first_arg = True
                        expected_args_counter = expected_args_counter + 1
                        yield result
                    else:
                        # print("NOT YET:", result)
                        result_buffer.append(result)
                        continue
                else:
                    if result["ARGS_COUNTER"] == expected_args_counter:
                        expected_args_counter = expected_args_counter + 1
                        yield result
                    else:
                        returned = getResult(result_buffer, expected_args_counter)
                        result_buffer.append(result)
                        if (
                            returned is not None
                            and returned["ARGS_COUNTER"] == expected_args_counter
                        ):
                            expected_args_counter = expected_args_counter + 1
                            yield returned

                        else:
                            continue

        else:
            # print(result)
            # print("Result buffer:", result_buffer)

            if len(result_buffer) >= 1:
                result_buffer = sorted(result_buffer, key=lambda k: k["ARGS_COUNTER"])
                for result in result_buffer:
                    yield result
                break
            else:
                break

    a.terminate()


def getResult(result_buffer, expected_args_counter):
    for result in result_buffer:
        if result["ARGS_COUNTER"] == expected_args_counter:
            del result_buffer[result_buffer.index(result)]
            return result
        else:
            pass


def map(
    func,
    args,
    callback=None,
    reducer=None,
    response_mode="OBJECT",
    chunksize=1,
    globals_dict=None,
    host=None,
    port=None,
    username=None,
    secret_key=None,
    runAchillesController=runAchillesController,
    abspath=abspath,
    dirname=dirname,
    multiprocess=multiprocess,
):

    if globals_dict is None:
        manager = multiprocess.Manager()

        globals_dict = {"OUTPUT_QUEUE": manager.Queue()}
    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        pass

    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )
            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")
            username = getenv("USERNAME")
            secret_key = getenv("SECRET_KEY")
        except (KeyError, NameError):
            host = host
            port = port
            username = username
            secret_key = secret_key

    a = multiprocess.Process(
        target=runAchillesController,
        args=(
            func,
            args,
            callback,
            reducer,
            response_mode,
            host,
            port,
            username,
            secret_key,
            globals_dict,
            chunksize,
        ),
    )

    a.start()
    while True:
        final_result = globals_dict["OUTPUT_QUEUE"].get()
        if final_result is not None:
            # print("FINAL_RESULT:", final_result)
            a.terminate()
            return final_result
        else:
            # print(final_result)
            time.sleep(1)


def setupGlobals():
    manager = multiprocess.Manager()

    globals_dict = {"OUTPUT_QUEUE": manager.Queue()}
    return globals_dict


def killCluster(
    command="KILL_CLUSTER",
    host=None,
    port=None,
    username=None,
    secret_key=None,
    command_verified=False,
    runAchillesController=runAchillesController,
):

    if (
        host is not None
        and port is not None
        and username is not None
        and secret_key is not None
    ):
        runAchillesController(
            command=command,
            host=host,
            port=port,
            username=username,
            secret_key=secret_key,
            command_verified=command_verified,
        )
    else:
        runAchillesController(command=command, command_verified=command_verified)


if __name__ == "__main__":
    runAchillesController()



#!/usr/bin/env python3

from sys import path
from os import getenv
from os.path import dirname, abspath, join
from dotenv import load_dotenv
from datetime import datetime

from multiprocess import Pool, cpu_count
import socket
import dill

from twisted.protocols.basic import LineReceiver
from twisted.internet.endpoints import TCP4ClientEndpoint, connectProtocol
from twisted.internet import reactor


class AchillesNode(LineReceiver):
    MAX_LENGTH = 999999999999999999999999999

    def __init__(self, host, port):

        self.HOST = host  # The server's hostname or IP address
        self.PORT = port  # The port used by the server
        self.connected = False
        self.client_id = -1
        self.func = None
        self.callback = None
        self.reducer = None

    def lineReceived(self, data):
        self.handleData(data)

    def handleData(self, data):
        data = dill.loads(data)
        if "GREETING" in data:
            greeting = data["GREETING"]
            client_id = data["CLIENT_ID"]
            self.client_id = client_id
            print(
                "GREETING:",
                f"{greeting}\nConnected to achilles_server running at {self.HOST}:{self.PORT}",
            )
            print("CLIENT_ID:", client_id)
            packet = dill.dumps(
                {
                    "IP": socket.gethostbyname(socket.gethostname()),
                    "CPU_COUNT": cpu_count(),
                    "DATETIME_CONNECTED": datetime.now(),
                    "CLIENT_ID": self.client_id,
                }
            )
            self.sendLine(packet)
        elif "START_JOB" in data:
            func = data["FUNC"]
            self.func = func
            callback = data["CALLBACK"]
            self.callback = callback
            reducer = data["REDUCER"]
            self.reducer = reducer
            packet = dill.dumps({"CLIENT_ID": self.client_id, "READY": True})
            # print("START_JOB RESPONSE:", packet)
            self.sendLine(packet)
        elif "ARG" in data:
            # print("ARG:", data["ARG"])
            with Pool(cpu_count()) as p:
                result = p.map(self.func, data["ARG"])
                if self.callback is not None:
                    result = p.map(self.callback, result)
                else:
                    pass
                p.close()
            if self.reducer is not None:
                result = self.reducer(result)
            else:
                pass
            packet = dill.dumps(
                {"ARGS_COUNTER": data["ARGS_COUNTER"], "RESULT": result}
            )
            # print("RESPONSE PACKET:", packet)
            self.sendLine(packet)
        elif "KILL_NODE" in data:
            self.sendLine(dill.dumps({"KILLED_CLUSTER": "KILLED_CLUSTER"}))
            self.transport.loseConnection()
            reactor.stop()


def runAchillesNode(host=None, port=None):
    if host is not None and port is not None:
        pass
    else:
        try:
            if __name__ != "__main__":
                import achilles

                dotenv_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
                )

                achilles_function_path = (
                    abspath(dirname(achilles.__file__)) + "\\lineReceiver\\"
                )
                path.append(achilles_function_path)

            else:
                basedir = abspath(dirname(__file__))
                dotenv_path = join(basedir, ".env")

                achilles_function_path = abspath(dirname(__file__))
                path.append(achilles_function_path)
            load_dotenv(dotenv_path, override=True)
            port = int(getenv("PORT"))
            host = getenv("HOST")

        except BaseException as e:
            print(f"No .env configuration file found ({e})...")
            host, port = genConfig()

    endpoint = TCP4ClientEndpoint(reactor, host, port)
    d = connectProtocol(endpoint, AchillesNode(host, port))

    reactor.run()


def genConfig(host=None, port=None):
    if __name__ != "__main__":
        import achilles

        dotenv_path = abspath(dirname(achilles.__file__)) + "\\lineReceiver\\.env"
    else:
        basedir = abspath(dirname(__file__))
        dotenv_path = join(basedir, ".env")
    if host is not None and port is not None:
        pass
    else:
        host = input("Enter HOST IP address:\t")
        port = int(input("Enter HOST port to connect to:\t"))
    with open(dotenv_path, "w") as config_file:
        config_file.writelines(f"HOST={host}\n")
        config_file.writelines(f"PORT={port}\n")
        config_file.close()
        print(
            f"Successfully generated .env configuration file at {dotenv_path}. Use achilles_node.genConfig() to overwrite."
        )
    return host, port


if __name__ == "__main__":
    runAchillesNode()



from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import build_ext

import os
import shutil

path = os.path.abspath(os.path.join(os.getcwd(), os.pardir))

achilles_modules = [
    "achilles_server.py",
    "achilles_node.py",
    "achilles_node2.py",
    "achilles_node3.py",
    "achilles_node4.py",
    "achilles_controller.py",
    "achilles_main.py",
]

for module in achilles_modules:
    shutil.copy(f"{path}\\{module}", f"{path}\\cythonized\\pyx\\{module}x")

ext_modules = [
    Extension("achilles_controller", ["pyx\\achilles_controller.pyx"]),
    Extension("achilles_main", ["pyx\\achilles_main.pyx"]),
    Extension("achilles_node", ["pyx\\achilles_node.pyx"]),
    Extension("achilles_node2", ["pyx\\achilles_node2.pyx"]),
    Extension("achilles_node3", ["pyx\\achilles_node3.pyx"]),
    Extension("achilles_node4", ["pyx\\achilles_node4.pyx"]),
    Extension("achilles_server", ["pyx\\achilles_server.pyx"]),
]

setup(name="achilles", cmdclass={"build_ext": build_ext}, ext_modules=ext_modules)



#!/usr/bin/env python3

import socket
import dill
from sys import stderr, path
import sqlite3
import json

from dotenv import load_dotenv
import yaml
from os import getenv
from os.path import abspath, dirname, join
import getpass

from twisted.protocols.basic import LineReceiver
from twisted.internet.endpoints import TCP4ClientEndpoint, connectProtocol
from twisted.internet import reactor
from multiprocess import cpu_count, Process, Queue, Manager
from datetime import datetime


class AchillesController(LineReceiver):
    MAX_LENGTH = 999999999999999999999999999999999

    def __init__(
        self,
        host,
        port,
        username,
        secret_key,
        achilles_function=None,
        achilles_args=None,
        achilles_callback=None,
        achilles_reducer=None,
        response_mode="OBJECT",
        globals_dict=None,
        chunksize=1,
        command=None,
        command_verified=False,
    ):

        self.HOST = host  # The server's hostname or IP address
        self.PORT = port  # The port used by the server
        self.USERNAME = username
        self.SECRET_KEY = secret_key
        self.response_mode = response_mode
        self.sqlite_db_created = False
        self.sqlite_db = ""
        self.abs_counter = 0
        self.achilles_function = achilles_function
        self.achilles_args = achilles_args
        self.achilles_callback = achilles_callback
        self.achilles_reducer = achilles_reducer
        self.globals_dict = globals_dict
        self.chunksize = chunksize
        self.command = command
        self.command_verified = command_verified

    def lineReceived(self, data):
        data = dill.loads(data)
        if "GREETING" in data:
            greeting = data["GREETING"]
            # print("GREETING:", greeting)
            packet = dill.dumps(
                {
                    "IP": socket.gethostbyname(socket.gethostname()),
                    "CPU_COUNT": cpu_count(),
                    "DATETIME_CONNECTED": datetime.now(),
                    "USERNAME": self.USERNAME,
                    "SECRET_KEY": self.SECRET_KEY,
                }
            )
            self.sendLine(packet)

        elif "AUTHENTICATED" in data:
            if data["AUTHENTICATED"] is True:
                print(
                    f"ALERT: Connection to achilles_server at {self.HOST}:{self.PORT} and authentication successful.\n"
                )
                if self.command == "KILL_CLUSTER":
                    self.kill_cluster()
                elif self.achilles_function is None and self.achilles_args is None:
                    self.command_interface()
                else:
                    self.init_achilles_compute()
            else:
                stderr.write("WARNING: Authentication failed.")
                self.transport.loseConnection()

        elif "PROCEED" in data:
            if self.achilles_function is None and self.achilles_args is None:
                proceed = input("Press ENTER if ready to proceed:\t")
                if proceed == "":
                    print(
                        "PROCEEDING WITH DISTRIBUTING ARGUMENTS AMONGST THE CONNECTED NODES..."
                    )
                    self.sendLine(dill.dumps({"VERIFY": True}))
                else:
                    stderr.write("ALERT: Job cancelled.")
                    self.command_interface()
            else:
                self.sendLine(dill.dumps({"VERIFY": True}))

        elif "RESULT" in data and self.response_mode == "STREAM":
            if self.achilles_function is None and self.achilles_args is None:
                # If STREAM is your chosen response mode, handle results packets here.
                print(data)
            else:
                self.globals_dict["OUTPUT_QUEUE"].put(data)

        elif "RESULT" in data and self.response_mode == "SQLITE":
            if self.sqlite_db_created is False:
                self.sqlite_db = datetime.now().strftime("%Y%m%d %H%M%S") + ".db"
                conn = sqlite3.connect(self.sqlite_db)
                c = conn.cursor()
                c.execute(
                    "CREATE TABLE results (args_counter real, abs_counter real, result text)"
                )
                self.sqlite_db_created = True
                if len(data["RESULT"]) == 1:
                    try:
                        instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {data['RESULT'][0]})"""
                        # print(instruction)
                        c.execute(instruction)
                        self.abs_counter = self.abs_counter + 1
                    except BaseException:
                        instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {json.dumps(data['RESULT'][0])})"""
                        # print(instruction)
                        c.execute(instruction)
                        self.abs_counter = self.abs_counter + 1

                else:
                    for i in range(len(data["RESULT"])):
                        try:
                            instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {data['RESULT'][i]})"""
                            # print(instruction)
                            c.execute(instruction)
                            self.abs_counter = self.abs_counter + 1
                        except BaseException:
                            instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {json.dumps(data['RESULT'][i])})"""
                            # print(instruction)
                            c.execute(instruction)
                            self.abs_counter + self.abs_counter + 1

                c.close()

            else:
                conn = sqlite3.connect(self.sqlite_db)
                c = conn.cursor()
                if len(data["RESULT"]) == 1:
                    try:
                        instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {data['RESULT'][0]})"""
                        # print(instruction)
                        c.execute(instruction)
                        self.abs_counter = self.abs_counter + 1
                    except BaseException:
                        instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {json.dumps(data['RESULT'][0])})"""
                        # print(instruction)
                        c.execute(instruction)
                        self.abs_counter = self.abs_counter + 1

                else:
                    for i in range(len(data["RESULT"])):
                        try:
                            instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {data['RESULT'][i]})"""
                            # print(instruction)
                            c.execute(instruction)
                            self.abs_counter = self.abs_counter + 1
                        except BaseException:
                            instruction = f"""INSERT INTO results VALUES ({data['ARGS_COUNTER']}, {self.abs_counter}, {json.dumps(data['RESULT'][i])})"""
                            # print(instruction)
                            c.execute(instruction)
                            self.abs_counter = self.abs_counter + 1

                c.close()

        elif "FINAL_RESULT" in data:
            if self.achilles_function is None and self.achilles_args is None:
                print("FINAL RESULT:", data["FINAL_RESULT"])
                self.command_interface()
            else:
                # print("FINAL RESULT:", data["FINAL_RESULT"])
                self.globals_dict["OUTPUT_QUEUE"].put(data["FINAL_RESULT"])
                self.transport.loseConnection()
                reactor.stop()

        elif "JOB_FINISHED" in data:
            if self.globals_dict is not None:
                self.globals_dict["OUTPUT_QUEUE"].put("JOB_FINISHED")
            else:
                self.transport.loseConnection()
                reactor.crash()

            self.transport.loseConnection()

        elif "CLUSTER_STATUS" in data:
            print("CLUSTER STATUS:", data)
            self.command_interface()

        elif "KILL_NODE" in data:
            stderr.write(
                "ALERT: All achilles_nodes have been disconnected from the cluster. The achilles_server is still running and accepting connections."
            )
            reactor.crash()

        else:
            print(data)
            self.command_interface()

    def achilles_compute(
        self, achilles_config_path=None, response_mode="", chunksize=1
    ):
        if __name__ != "__main__":
            import achilles

            achilles_function_path = (
                abspath(dirname(achilles.__file__)) + "\\lineReceiver\\"
            )
            # print(achilles_function_path)
            path.append(achilles_function_path)
        else:
            achilles_function_path = abspath(dirname(__file__))
            # print(achilles_function_path)
            path.append(achilles_function_path)

        if self.achilles_function is None and self.achilles_args is None:
            from achilles.lineReceiver.achilles_function import (
                achilles_function,
                achilles_args,
                achilles_callback,
                achilles_reducer,
            )

            achilles_config_path = f"{achilles_function_path}\\{achilles_config_path}"
            with open(achilles_config_path, "r") as f:
                achilles_config = yaml.load(f, Loader=yaml.Loader)
                try:
                    args_path = (
                        f"{achilles_function_path}\\{achilles_config['ARGS_PATH']}"
                    )
                except KeyError:
                    args_path = None
                try:
                    achilles_args = achilles_config["ARGS"]
                except KeyError:
                    achilles_args = achilles_args
                modules = achilles_config["MODULES"]
                group = achilles_config["GROUP"]

        else:
            achilles_function = self.achilles_function
            achilles_args = self.achilles_args
            achilles_callback = self.achilles_callback
            achilles_reducer = self.achilles_reducer
            args_path = None
            modules = None
            group = None

        self.response_mode = response_mode

        packet = dill.dumps(
            {
                "FUNC": achilles_function,
                "ARGS": achilles_args,
                "ARGS_PATH": args_path,
                "MODULES": modules,
                "CALLBACK": achilles_callback,
                "REDUCER": achilles_reducer,
                "GROUP": group,
                "RESPONSE_MODE": response_mode,
                "CHUNKSIZE": chunksize,
            }
        )
        self.sendLine(packet)

    def get_cluster_status(self):
        packet = dill.dumps({"GET_CLUSTER_STATUS": "GET_CLUSTER_STATUS"})
        self.sendLine(packet)
        print("ALERT: Requested cluster status.")

    def kill_cluster(self):
        if self.command_verified is True:
            packet = dill.dumps({"KILL_CLUSTER": "KILL_CLUSTER"})
            self.sendLine(packet)
        else:
            confirm_kill_cluster = input(
                "WARNING: Are you absolutely sure that you want to kill the cluster? Enter YES to proceed:\t"
            )
            if confirm_kill_cluster == "YES":
                packet = dill.dumps({"KILL_CLUSTER": "KILL_CLUSTER"})
                self.sendLine(packet)
            else:
                stderr.write("ALERT: kill_cluster aborted.\n")
                self.command_interface()

    def command_interface(self):
        command = input("Achilles cluster is ready to accept commands:\t")
        if command == "achilles_compute":
            self.init_achilles_compute()
        elif command == "cluster_status":
            self.get_cluster_status()
        elif command == "kill_cluster":
            self.kill_cluster()
        elif command == "help":
            print("\n-------\nCommands:")
            print("achilles_compute, cluster_status, kill_cluster, help\n-------\n")
            self.command_interface()
        else:
            stderr.write(
                "Sorry, that command is not recognized. Type 'help' for a list of commands.\n"
            )
            self.command_interface()

    def init_achilles_compute(self):
        if self.achilles_function is None and self.achilles_args is None:
            achilles_config_path = input(
                "Enter path to achilles_config.yaml to begin job:\t"
            )
            self.response_mode = input(
                f"Enter desired response mode (OBJECT, SQLITE, or STREAM):\t"
            )
        else:
            achilles_config_path = None
        if self.response_mode in ["OBJECT", "SQLITE", "STREAM"]:
            self.achilles_compute(
                achilles_config_path=achilles_config_path,
                response_mode=self.response_mode,
                chunksize=self.chunksize,
            )
        else:
            stderr.write(
                "Sorry, that response mode is not recognized. Please choose OBJECT, SQLITE or STREAM.\n"
            )
            self.init_achilles_compute()



#!/usr/bin/env python3

# from requests import get
# from html2text import html2text


def achilles_args(args_path):
    import ast

    with open(args_path, "r") as args:
        for arg in args:
            arg = arg.strip()
            print(ast.literal_eval(arg))
            for a in ast.literal_eval(arg):
                print(a)
            yield ast.literal_eval(arg)


def achilles_function(arg):
    return arg ** 2


def achilles_callback(result):
    return result ** 2


def achilles_reducer(list_of_results):
    sum_of_results = [0]
    for result in list_of_results:
        sum_of_results[0] += result
    return sum_of_results


"""def achilles_function2(arg):
    r = get(
        arg,
        headers={"User-Agent": "Mozilla/5.0", "referrer": "https://l.facebook.com/"},
    )
    raw_html = r.text
    markdown_text = html2text(raw_html)

    return markdown_text"""
